SET NOCOUNT ON;
go
USE ImEx;
go
IF EXISTS
 (SELECT *
  FROM   sys.objects
  WHERE  object_id = OBJECT_ID(N'dbo.SetBasedIter1BMin25')
  AND    type IN (N'P', N'PC'))
BEGIN;
  DROP PROCEDURE dbo.SetBasedIter1BMin25;
END;
go
CREATE PROCEDURE dbo.SetBasedIter1BMin25
                 @MaxCandidatesPerSession smallint
AS
-- Create temporary table for efficient update of SpaceLeft in dbo.Sessions
CREATE TABLE #SessionsAssigned
            (Year          smallint NOT NULL
            ,Quarter       tinyint  NOT NULL
            ,SessionNo     smallint NOT NULL
            ,NumCandidates smallint NOT NULL
            ,CONSTRAINT PK_SessionsAssigned
                 PRIMARY KEY (Year, Quarter, SessionNo)
            );

-- For each quarter, compute the minimal number of sessions needed for the registrations.
-- Insert that number of (empty) sessions in the table.
WITH SessionsNeeded
AS (SELECT   Year, Quarter,
             CEILING(0.75 * SUM(NumCandidates) / @MaxCandidatesPerSession) AS Needed
    FROM     dbo.Registrations
    GROUP BY Year, Quarter)
INSERT INTO  dbo.Sessions (Year, Quarter, SessionNo, SpaceLeft)
SELECT       sn.Year,
             sn.Quarter,
             n.Number,
             @MaxCandidatesPerSession
FROM         SessionsNeeded AS sn
INNER JOIN   dbo.Numbers    AS n
      ON     n.Number  BETWEEN 1 AND sn.Needed;

-- Repeat outer loop until no more sessions are added before loop or in last statement of loop
WHILE @@ROWCOUNT <> 0
BEGIN;

  -- Add one entry to each new (empty) session.
  -- NOTE: Test for empty session and unassigned registration is still relevant - this may be the first
  --       execution of this query in the inner loop, but it need not be the first iteration of the outer loop.
  WITH RankedSessions
  AS (SELECT     Year, Quarter, SessionNo, SpaceLeft,
                 ROW_NUMBER()     OVER (PARTITION BY Year, Quarter
                                        ORDER BY (SELECT 1))        -- Actual ranking irrelevant, but ORDER BY is mandatory
                                     AS Ranking
      FROM       dbo.Sessions
      WHERE      SpaceLeft            = @MaxCandidatesPerSession)   -- Only new sessions for the first iteration
 ,RankedRegistrations
  AS (SELECT     r.Year, r.Quarter, r.NumCandidates, r.SessionNo,
                 ROW_NUMBER()     OVER (PARTITION BY r.Year, r.Quarter
                                        ORDER BY r.NumCandidates DESC)
                                     AS Ranking
      FROM       dbo.Registrations   AS r
      WHERE      r.SessionNo         IS NULL)                       -- Still relevant (may be first iteration for second set of sessions)
  UPDATE         r
  SET            SessionNo            = s.SessionNo
  OUTPUT         inserted.Year,         inserted.Quarter,
                 inserted.SessionNo,    inserted.NumCandidates
  INTO           #SessionsAssigned (Year, Quarter, SessionNo, NumCandidates)
  FROM           RankedRegistrations AS r
  INNER JOIN     RankedSessions      AS s
        ON       s.Year               = r.Year
        AND      s.Quarter            = r.Quarter
        AND      s.Ranking            = r.Ranking;

  -- Repeat inner loop until no more registrations handled before loop (can't happen) or in last statement of loop
  WHILE @@ROWCOUNT > 0
  BEGIN;

    -- One or more registrations were assigned to sessions.
    -- Subtract their size from remaining room in those sessions.
    UPDATE         s
    SET            s.SpaceLeft          = s.SpaceLeft - sa.NumCandidates
    FROM           dbo.Sessions        AS s
    INNER JOIN     #SessionsAssigned   AS sa
          ON       sa.Year              = s.Year
          AND      sa.Quarter           = s.Quarter
          AND      sa.SessionNo         = s.SessionNo;

    -- Clear out the temp table used for updating SpaceLeft after assigning registrations to sessions
    TRUNCATE TABLE #SessionsAssigned;
    
    -- Add another entry to each session that still has room for one.
    WITH Threshold
    AS (SELECT     Year, Quarter,
                   MAX(SpaceLeft)      AS Threshold
        FROM       dbo.Sessions
        GROUP BY   Year, Quarter)
   ,RankedSessions
    AS (SELECT     Year, Quarter, SessionNo, SpaceLeft,
                   ROW_NUMBER()     OVER (PARTITION BY Year, Quarter
                                          ORDER BY SpaceLeft DESC)
                                       AS Ranking
        FROM       dbo.Sessions
        WHERE      SpaceLeft            > 0)
   ,RankedRegistrations
    AS (SELECT     r.Year, r.Quarter, r.NumCandidates, r.SessionNo,
                   ROW_NUMBER()     OVER (PARTITION BY r.Year, r.Quarter
                                          ORDER BY r.NumCandidates DESC)
                                       AS Ranking
        FROM       dbo.Registrations   AS r
        INNER JOIN Threshold           AS t
              ON   t.Year               = r.Year
              AND  t.Quarter            = r.Quarter
        WHERE      r.SessionNo         IS NULL
        AND        r.NumCandidates     <= t.Threshold)
    UPDATE         r
    SET            SessionNo            = s.SessionNo
    OUTPUT         inserted.Year,         inserted.Quarter,
                   inserted.SessionNo,    inserted.NumCandidates
    INTO           #SessionsAssigned (Year, Quarter, SessionNo, NumCandidates)
    FROM           RankedRegistrations AS r
    INNER JOIN     RankedSessions      AS s
          ON       s.Year               = r.Year
          AND      s.Quarter            = r.Quarter
          AND      s.Ranking            = r.Ranking
    WHERE          s.SpaceLeft         >= r.NumCandidates;

  END; -- End inner loop (adding registrations to sessions)

  -- For each quarter, compute the minimal number of addition sessions needed for the unallotted registrations.
  -- Insert that number of (empty) sessions in the table.
  WITH SessionsNeeded
  AS (SELECT   Year, Quarter,
               CEILING(0.75 * SUM(NumCandidates) / @MaxCandidatesPerSession) AS Needed
      FROM     dbo.Registrations
      WHERE    SessionNo IS NULL         -- Exlude registrations that are already allocated
      GROUP BY Year, Quarter)
 ,MaxSession
  AS (SELECT   Year, Quarter,
               MAX(SessionNo) AS MaxSession
      FROM     dbo.Sessions
      GROUP BY Year, Quarter)
  INSERT INTO  dbo.Sessions (Year, Quarter, SessionNo, SpaceLeft)
  SELECT       sn.Year,
               sn.Quarter,
               COALESCE(ms.MaxSession, 0) + n.Number,
               @MaxCandidatesPerSession
  FROM         SessionsNeeded AS sn
  INNER JOIN   dbo.Numbers    AS n
        ON     n.Number  BETWEEN 1 AND sn.Needed
  INNER JOIN   MaxSession     AS ms
        ON     ms.Year         = sn.Year
        AND    ms.Quarter      = sn.Quarter;

END; -- End outer loop (adding sessions to be filled in inner loop)
go
